#!/usr/bin/env python

import argparse
import json
import os
import sys
import textwrap

from redis import StrictRedis
from plurmy import Slurm

def parse_args(args):
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest='command')
    stage_doc = textwrap.dedent("""\

    The stage subcommand pushes messages to a user defined redis queue.
    The messages can be any serializable message, such as a file
    path, a WKT string, a PVL label, or a JSON message.

    Messages are staged using the redis pipeline capability so large
    numbers of messages (> 100k for example), can be efficiently pushed
    to the redis queue.
    
    The stage subcommand appends ':processing' to the queue name in
    order to make use of atomic push/pop operations if and when
    the process subcommand is executed. If one wishes to double check
    that messages have been pushed to their processing queue, it is necessary
    to append ':processing' to the queue name. See the example below.

    Usage Examples
    --------------
    $ ls *.img > mylist.lis
    $ plurmy stage <queuename> mylist.lis
      Pushed X messags to <queuename>:processing.
    $ redis-cli llen <queuename>:processing
      X (where X is the number of lines in mylist.lis)
    """)
    stage = subparsers.add_parser('stage', 
                                  help='Stage jobs to a processing queue', 
                                  formatter_class=argparse.RawDescriptionHelpFormatter,
                                  epilog=stage_doc)
    stage.add_argument('-p', '--port', help='The redis port. Alternatively, set the PLURMY_REDIS_PORT env variable', type=int)
    stage.add_argument('-o', '--host', help='The redis host to connect to. Alternatively, set the PLURMY_REDIS_HOST env variable.', default=None)
    stage.add_argument('queue', help='The name of the queue to stage work to.')
    stage.add_argument('inputfile', help='A text file delimited by a new line character containing the file to stage.')

    restage_doc = textwrap.dedent("""\
    This command moves messages from a <queuename>:working queue 
    back to the <queuename>:processing queue.
    
    When processing jobs from a processing queue on a large number
    of cores, it is highly likely that jobs will fail due to any 
    number of reasons. The plurmy command works by staging 
    messages to a processing queue and then atomically (without an
    opportunity for message lose) pushing those message to a working
    queue while processing is underway. At the conclusion of a successful
    job run, the message is removed (lrem) from the working queue. At 
    the conclusion of a long processing run with many jobs some number
    of jobs could remain on the working queue. 

    After messages have been restaged, it is highly likely that another
    processing run will be submitted using plurmy process.
    
    Usage Examples
    --------------
    $ plurmy restage <queuename>
      Moved X messages from <queuename>:working to <queuename>:processing
    """)

    restage = subparsers.add_parser('restage', 
                                    help='Restage failed jobs from a working queue to a processing queue.',
                                    formatter_class=argparse.RawDescriptionHelpFormatter,
                                    epilog=restage_doc)
    restage.add_argument('queue', help='The name of the queue to stage work to.')
    restage.add_argument('-p', '--port', help='The redis port. Alternatively, set the PLURMY_REDIS_PORT env variable', type=int)
    restage.add_argument('-o', '--host', help='The redis host to connect to. Alternatively, set the PLURMY_REDIS_HOST env variable.', default=None)
    process_doc = textwrap.dedent("""\
    The process subcommand is used to submit jobs that draw messages from a redis
    queue. The process command wraps a user provided command (a script or a 
    shell command most likely) and appends the message from the queue as the
    final argument.

    The script passed to the process subcommand (cmd) should be quoted
    as it is passed as a simple string. By default, process sets the number
    of tasks to one (-n=1). The command passed will be executed with 'srun -n 1 <cmd>'
    meaning that a single, non-parallel job will be executed, drawing messages
    from the processing queue. If the number of specified tasks is greater than one
    srun will execute n concurrent, tasks. For example, if -n=40, plurmy process
    will request 40 CPUs and execute 'srun -n 40 <cmd>'. A logic upper limit for n is the
    number of cores available on the cluster (Be kind to your neighbors...).

    The jobs submitted by plurmy process wrap the user supplied command in the following
    block:

    ```
    srun -n {args.ntasks} sh << EOF
    while :
      do
        msg=$(redis-cli -h {args.host} -p {args.port} rpoplpush {processing_queue_name} {working_queue_name})

        # When the queue is empty, break
        if [ -z "$msg" ]; 
        then 
          break 
        fi

        # Execute the passed command(s) 
        
        if . {args.cmd} $msg; then
            redis-cli -h ${args.host} -p ${args.port} lrem {working_queue_name} -1 "$msg"
        fi
      done
    EOF
    ```

    This block reads messages from the user supplied queue (e.g, do_things), atomically pushes the
    message to a working queue (e.g., do_things:working), executes the user supplied command, 
    and then, if successful, removes the message from the working queue.

    Assumptions
    -----------
    - the final positional argument passed to the command is the message
    - the command provides a non-zero exit code if it fails to execute properly
    - the command is wrapped in single quotes on submission as the parser parses it as a simple string
    - the queue is specified by the base name, e.g., my_queue and not my_queue:processing, the
      script appends :processing and :working to the user supplied queue name

    Usage Examples
    --------------
    In this first example 40 CPUs are requested with a 10 minute walltime and 4GB of RAM.
    Under this allocation, an srun command is submitted to run 40 concurrent tasks.
    
    plurmy process -j myjob -m 4000 -t '00:10:00' -n 40 myqueue 'my_awesome_script.sh <arg1> <arg2>'
    """)

    process = subparsers.add_parser('process', 
                                    help='Submit jobs to a cluster that draw from a queue.',
                                    formatter_class=argparse.RawDescriptionHelpFormatter,
                                    epilog=process_doc)
    process.add_argument('queue', help='The processing queue to draw messages from.')
    process.add_argument('cmd', help="The single quoted command to execute on each staged file.")
    process.add_argument('--notbashscript', action='store_true',
                         help='A flag to set if the script being passed is not a bash script.')
    process.add_argument('-e', '--env', help='An optional conda environment to activate', default=None)
    process.add_argument('-j', '--jobname', help='The name of the job.', default='plurmy')
    process.add_argument('-m', '--mem-per-cpu', help='The amount of RAM to request (in MB) per CPU.', 
                         default=2000, type=int)
    process.add_argument('-l', '--log-dir', help='Directory to place output logs.')
    process.add_argument('-n', '--ntasks', help='The number of CPUs to request.',
                         default=1, type=int)
    process.add_argument('-t', '--time', help='The amount of walltime to request.',
                        default="00:60:00")
    process.add_argument('--time-min', help='The minimum amount of walltime to run the job.',
                         default="00:30:00")
    process.add_argument('-p', '--port', help='The redis port. Alternatively, set the PLURMY_REDIS_PORT env variable', type=int)
    process.add_argument('-o', '--host', help='The redis host to connect to. Alternatively, set the PLURMY_REDIS_HOST env variable.', default=None)
    process.add_argument('-q', '--partition', help='The slurm processing queue to submit jobs to.')
    return parser, parser.parse_args(args)

def setup_redis_connection(host: str, port: int):
    """
    Setup a redis connection to the provided host and port. Default
    to using database 0.

    Parameters
    ----------
    host : str
           The IP or hostname where redis is running

    port : int
           The port that redis is listening at

    Returns
    -------
     : object
       A redis-py queue object
    """
    return StrictRedis(host=host, port=port, db=0)

def read_file(inputfile: str):
    """
    Read the lines of an input file into a list, running
    rstrip() on each line.

    Parameters
    ----------
    inputfile : str
                Path to the input file

    Returns
    -------
    iterable : list
               of strings where each entry is a line in the 
               input file
    """
    with open(inputfile, 'r') as stream:
        iterable = stream.readlines()
        iterable = [i.rstrip() for i in iterable]
    return iterable

def stage(redis_queue, iterable: list, processing_queue_name: str):
    """
    Stage an iterable of messages to a redis queue.

    Parameters
    ----------
    redis_ queue : object
                  A redis-py redis queue object

    iterable : list
               of serialized messages to push to the queue

    processing_queue_name : str
                            The name of the queue to pass the messages to

    """
    pipeline = redis_queue.pipeline()
    for i, msg in enumerate(iterable):
        pipeline.rpush(processing_queue_name, msg)
    pipeline.execute()
    print(f'Pushed {i+1} messages to {processing_queue_name}.')

def process(args, processing_queue_name: str, working_queue_name: str):
    """
    Submit jobs to a Slurm controller using sbatch directives to process data
    from a given processing queue.

    This func handles atomically pushing from the processing queue to the working
    queue and then popping messages off the working queue upon sucessful 
    completion.

    Parameters
    ----------
    args : object
           An argparse namespace with arguments to be parsed for job creation

    processing_queue_name : str
                            The name of the processing queue to draw work from

    working_queue_name : str
                         The name of the queue to push messages to while work is
                         underway.
    """

    if args.notbashscript == False:
      argscmd = f'. {args.cmd}'
    else:
      argscmd = args.cmd

    if args.env:
        env = f'conda activate --stack {args.env}'
    else:
        env = ''
    conda_init = os.environ.get('PLURMY_CONDA_INIT', None)
    if not conda_init:
        raise EnvironmentError('"PLURMY_CONDA_INIT" must be set to the path of the conda.sh script, e.g., /home/user/anaconda3/etc/profile.d/conda.sh')

    cmd = textwrap.dedent(f"""\
    srun bash <<'EOF'
    source {conda_init}
    conda activate plurmy
    {env}
    while :
      do
        msg=$(redis-cli -h {args.host} -p {args.port} rpoplpush {processing_queue_name} {working_queue_name})
        # When the queue is empty, break
        if [ -z "$msg" ]; then 
          break 
        fi

        # Execute the passed command(s); since this is executed in a subshell, the script could just use $msg. Explixit passing is preferred.
        
        if {argscmd} $msg; then
            redis-cli -h {args.host} -p {args.port} lrem {working_queue_name} -1 "$msg" > /dev/null
        fi
      done
    EOF
    """)

    submitter = Slurm(cmd,
                      job_name=args.jobname,
                      mem_per_cpu=args.mem_per_cpu,
                      time=args.time,
                      ntasks=args.ntasks,
                      output=args.log_dir,
                      account='astro',
                      partition=args.partition)
    job_str = submitter.submit()
    print(job_str)

def restage(redis_queue, processing_queue_name: str, working_queue_name: str):
    """
    Restage work from a working queue back onto a processing queue.

    Parameters
    ----------
    redis_ queue : object
                A redis-py redis queue object
   
    processing_queue_name : str
                            The name of the processing queue to draw work from

    working_queue_name : str
                         The name of the queue to push messages to while work is
                         underway. For restage these are, presumably, failed jobs
    """

    pipeline = redis_queue.pipeline()

    n_messages = redis_queue.llen(working_queue_name)
    for _ in range(n_messages):
        pipeline.rpoplpush(working_queue_name, processing_queue_name)
    pipeline.execute()
    print(f'Moved {n_messages} from {working_queue_name} to {processing_queue_name}.')

def get_queue_names(queue: str):
    """
    Generate working and processing queue names

    Parameters
    ----------
    queue : str
            The name of the queue

    Returns
    -------
      :
       Two strings with ':processing' and ':working' appended to 
       the input queue name
    """
    return f'{queue}:processing', f'{queue}:working'

def get_redis_host_port(args):
  """
  Argument validation to override the host/port if provided as arguments
  to this script. If the host/port arguments are not overridden, get the
  values from the environment and raise an error if still None.

  Parameters
  ----------
  args : object
         An argparse Namespace

  Returns
  -------
  host : str
         The redis hostname
  
  port : int
         The port redis is listening on
  """
  if args.host is None:
    host = os.environ.get('PLURMY_REDIS_HOST', None)
  else:
    host = args.host
  if args.port is None:
    port = os.environ.get('PLURMY_REDIS_PORT', None)
  else:
    port = args.port

  if host is None:
    raise ValueError(f'Redis host is {host}. A valid redis host must be set using either the --host argument or the PLURMY_REDIS_HOST environment variable.')
  if port is None:
    raise ValueError(f'Redis port is {port}. A valid redis port must be set using either the --port argument or the PLURMY_REDIS_PORT environment variable.')

  return host, int(port)

def main(parser=None, args=None):
    """
    The dispatcher for the job commands

    Parameters
    ----------
    parser : object
             An argparse ArgumentParser object

    args : object
           An argparse Namespace
    """
    if parser == None:
        parser, args = parse_args(sys.argv[1:])
    processing_queue_name, working_queue_name = get_queue_names(args.queue)

    host, port = get_redis_host_port(args)

    if args.command == 'stage':
        redis_queue = setup_redis_connection(host=host, port=port)
        iterable = read_file(args.inputfile)
        stage(redis_queue, iterable, processing_queue_name)
    elif args.command =='restage':
        redis_queue = setup_redis_connection(host=host, port=port)
        restage(redis_queue, processing_queue_name, working_queue_name)
    elif args.command == 'process':
        process(args, processing_queue_name, working_queue_name)
    else:
        parser.print_help()

if __name__ == '__main__':
    main(parse_args(sys.argv[1:]))
 



from unittest import mock

import pytest
import plurmy


@pytest.fixture
def job():
    return plurmy.Slurm(
        "foo",
        job_name="SpecialJob",
        time="00:00:60",
        output="/mi/casa",
        mem_per_cpu="4096",
        partition="longall",
    )


@pytest.mark.parametrize(
    "time, expected",
    [
        ("1:00:00", 3600),
        ("01:00:00", 3600),
        ("00:01:00", 60),
        ("00:00:60", 60),
        ("00:00:90", 90),
    ],
)
def test_slurm_walltime_to_seconds(time, expected):
    print(plurmy.__file__)

    assert plurmy.slurm_walltime_to_seconds(time) == expected


@pytest.mark.parametrize(
    "command, output, err, kwargs, expected",
    [
        ("foo", "", None, {}, "foo"),  # Check command is passing
        ("foo", "", None, {"name": "SpecialJob"}, "SpecialJob"),
        ("foo", "", None, {"time": "00:00:60"}, "00:00:60"),
        ("foo", "", None, {"outdir": "/mi/casa"}, "/mi/casa"),
        ("foo", "", None, {"mem": 4096}, "4096"),
        ("foo", "", None, {"queue": "longall"}, "longall"),
        ("foo", "", None, {"ntasks": 3}, "--ntasks=3"),
        ("source activate krc\nfoo", "", None, {"env": "krc"}, "krc"),
        ("foo", "", "error", {}, "error"),
    ],
)
def test_spawn(job, command, output, err, kwargs, expected):
    job.command = command
    for k, v in kwargs.items():
        setattr(job, k, v)
    with mock.patch("subprocess.Popen") as mock_popen:
        mock_popen.return_value.communicate.return_value = (output, err)
        js = job.submit()
        if err is None:
            assert expected in js
        else:
            assert js is False


@pytest.mark.parametrize(
    "command, array, output, err, kwargs, expected",
    [
        ("foo", "1-2", "", None, {}, "foo"),  # Check command is passing
        ("foo", "1-50", "", None, {"outdir": "/mi/casa"}, "/mi/casa"),
        ("foo", "1-6", "", None, {"queue": "longall"}, "longall"),
        ("source activate krc\nfoo", "1-6", "", None, {"env": "krc"}, "krc"),
        ("foo", "1-2", "", "error", {}, "error"),
        ("foo", "1-2", "", None, {"spread_job": ""}, "spread-job"),
    ],
)
def test_simple_job_arr(job, command, array, output, err, kwargs, expected):
    for k, v in kwargs.items():
        setattr(job, k, v)

    with mock.patch("subprocess.Popen") as mock_popen:
        mock_popen.return_value.communicate.return_value = (output, err)

        js = job.submit(array=array)
        if err is None:
            call = mock.call(["sbatch", "--array", array], stdin=-1, stdout=-1)
            assert mock_popen.call_args_list == [call]
            assert len(js) == 1
            assert expected in js[0]
        else:
            assert js is False


@pytest.mark.parametrize(
    "command, array, output, err, nchunks, kwargs, expected",
    [
        ("foo", "1-400%2", "", None, 1, {"time": "00:00:60"}, "00:00:60"),
        ("foo", "1-1500", "", None, 2, {}, "foo"),
        ("foo", "1-1500%500", "", None, 2, {}, "foo"),
        ("foo", "1-2003%250", "", None, 3, {}, "foo"),
    ],
)
def test_chunked_job_arr(job, command, array, output, err, nchunks, kwargs, expected):
    for k, v in kwargs.items():
        setattr(job, k, v)

    with mock.patch("subprocess.Popen") as mock_popen:
        mock_popen.return_value.communicate.return_value = (output, err)

        js = job.submit(array=array)

        if err is None:
            assert len(js) == nchunks
            for i in range(nchunks):
                assert expected in js[i]

        else:
            assert js is False


@pytest.mark.parametrize(
    "command, array, output, err, kwargs, expected",
    [("foo", "1-3,5-8", "", None, {"name": "SpecialJob"}, "SpecialJob")],
)
def test_compound_job_array(job, command, array, output, err, kwargs, expected):

    for k, v in kwargs.items():
        setattr(job, k, v)

    with mock.patch("subprocess.Popen") as mock_popen:
        mock_popen.return_value.communicate.return_value = (output, err)

        js = job.submit(array=array)
        if err is None:
            calls = []

            for arr in array.split(","):
                calls.append(mock.call(["sbatch", "--array", arr], stdin=-1, stdout=-1))

            for i in range(len(calls)):
                assert mock_popen.call_args_list[i] == calls[i]
            for j in js:
                assert expected in j
        else:
            assert js is False



import argparse
from unittest.mock import mock_open, patch, create_autospec
import pytest
import os

from plurmy import orchestrate as orc

import fakeredis

@pytest.mark.parametrize('args, expected',[
    (['stage', 'my_queue', 'my_input'], {'command':'stage', 'queue':'my_queue', 'inputfile':'my_input', 
                                         'host':None, 'port':None}),
    (['restage', 'my_queue'], {'command':'restage', 'queue':'my_queue', 'port':None,
                                      'host':None}),
    (['process', 'my_queue', 'cmd'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}),
    (['process', 'my_queue', 'cmd', '-j', 'foo'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host':None, 'jobname':'foo',
                                      'notbashscript':False, 'env':None}),
    (['process', 'my_queue', 'cmd', '-m', '4000'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':4000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}),
    (['process', 'my_queue', 'cmd', '-l', '/here'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':'/here', 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host': None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}),
    (['process', 'my_queue', 'cmd', '-n', '40'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':40,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}),
    (['process', 'my_queue', 'cmd', '-t', '00:15:00'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:15:00', 'time_min':'00:30:00', 'port':None,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}), 
    (['process', 'my_queue', 'cmd', '--time-min', '00:60:00'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:60:00', 'port':None,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}), 
    (['process', 'my_queue', 'cmd', '--port', '4000'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':4000,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}), 
    (['process', 'my_queue', 'cmd', '--host', 'host'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host':'host', 'jobname':'plurmy',
                                      'notbashscript':False, 'env':None}),  
    (['process', 'my_queue', 'cmd', '--notbashscript'], {'command':'process', 'queue':'my_queue', 'cmd':'cmd',
                                      'mem_per_cpu':2000, 'log_dir':None, 'ntasks':1,
                                      'time':'00:60:00', 'time_min':'00:30:00', 'port':None,
                                      'host':None, 'jobname':'plurmy',
                                      'notbashscript':True, 'env':None}),                                                                                                                                                                                          
])
def test_parse_args(args, expected):
    parser, args = orc.parse_args(args)
    expected = argparse.Namespace(**expected)
    assert args == expected

@pytest.mark.parametrize('args, expected', [
    (['-h'], '{stage,restage,process}'),
    (['stage', '-h'], 'plurmy stage <queuename> mylist.lis'),
    (['restage', '-h'], 'plurmy restage <queuename>'),
    (['process', '-h'], 'plurmy process -j myjob -m 4000')
])
def test_print_help(capfd, args, expected):
    with pytest.raises(SystemExit):
        parser, args = orc.parse_args(args)
    out, err = capfd.readouterr()
    assert expected in out

def test_setup_redis_connection():
    def mock_redis_constructor(host=None, port=None, db=0):
        pass
    
    with patch('plurmy.orchestrate.StrictRedis', create_autospec(mock_redis_constructor)) as _conn:
        orc.setup_redis_connection('host', 12345)
        _conn.assert_called_once_with('host', 12345, db=0)

def test_readfile():
    with patch('plurmy.orchestrate.open', mock_open()) as _file:
        iterable = orc.read_file('foo.lis')
        _file.assert_called_once_with('foo.lis', 'r')
    
def test_stage():
    redis_queue = fakeredis.FakeStrictRedis()
    iterable = ['fee', 'fi', 'fo', 'fum']
    orc.stage(redis_queue, iterable, 'test_stage_queue')
    assert len(iterable) == redis_queue.llen('test_stage_queue')

    for i in range(len(iterable)):
        # Need to encode because redis strings are binary
        assert redis_queue.lpop('test_stage_queue') == iterable[i].encode()

@pytest.mark.parametrize('args, env, expected_host, expected_port', [
    (argparse.Namespace(host=None, port=None), {'PLURMY_REDIS_HOST':'localhost', 'PLURMY_REDIS_PORT':'1000'}, 'localhost', 1000),
    (argparse.Namespace(host='foo', port=None), {'PLURMY_REDIS_HOST':'notfoo', 'PLURMY_REDIS_PORT':'6379'}, 'foo', 6379),
    (argparse.Namespace(host=None, port=1000), {'PLURMY_REDIS_HOST':'localhost', 'PLURMY_REDIS_PORT':'6379'}, 'localhost', 1000)
])
def test_get_redis_host_port(args, env, expected_host, expected_port):
    
    with patch.dict(os.environ, env):
        host, port = orc.get_redis_host_port(args)
        assert host == expected_host
        assert port == expected_port

@pytest.mark.parametrize('args, env, errmsg', [
    (argparse.Namespace(host=None, port=None), {}, 'host'),
    (argparse.Namespace(host='foo', port=None), {}, 'port'),
])
def test_raise_redis_host_port(args, env, errmsg):
    with patch.dict(os.environ, env, clear=True):
        with pytest.raises(ValueError) as e:
            host, port = orc.get_redis_host_port(args)
        assert errmsg in str(e.value)

@pytest.mark.parametrize('args, cmd, env', [
    (argparse.Namespace(jobname='test', mem_per_cpu=2000, time='00:05:00', ntasks=1, 
                        log_dir=None, host='localhost', port=6547, cmd='my_custom_command',
                        notbashscript=False, env=None), '. my_custom_command', ''),
    (argparse.Namespace(jobname='test', mem_per_cpu=2000, time='00:05:00', ntasks=1, 
                        log_dir=None, host='localhost', port=6547, cmd='my_custom_command',
                        notbashscript=True, env=None), 'my_custom_command', ''),
    (argparse.Namespace(jobname='test', mem_per_cpu=2000, time='00:05:00', ntasks=1, 
                        log_dir=None, host='localhost', port=6547, cmd='my_custom_command',
                        notbashscript=True, env='/my/conda/env'), 'my_custom_command', 'conda activate --stack /my/conda/env')

])
def test_process(capfd, args, cmd, env):    
    with patch("subprocess.Popen") as mock_popen:
        mock_popen.return_value.communicate.return_value = ('', '')

        orc.process(args, 'processing_queue', 'working_queue')
    out, err = capfd.readouterr()
    assert 'if [ -z "$msg" ]; then' in out
    assert f'if {cmd} $msg; then' in out
    assert 'srun' in out
    assert env in out

def test_restage():
    redis_queue = fakeredis.FakeStrictRedis()
    working_queue_name = 'test_restage:working'
    processing_queue_name = 'test_restage:processing'
    redis_queue.rpush(working_queue_name, 'Fake Message')
    assert redis_queue.llen(working_queue_name) == 1

    orc.restage(redis_queue, processing_queue_name, working_queue_name)

    assert redis_queue.llen(working_queue_name) == 0
    assert redis_queue.llen(processing_queue_name) == 1

def test_get_queue_names():
    processing, working = orc.get_queue_names('queue')
    assert processing == 'queue:processing'
    assert working == 'queue:working'

@pytest.mark.parametrize('cmd', [
    ('stage'),
    ('restage'),
    ('process')
])
def test_main(capfd, cmd):
    parser = argparse.ArgumentParser()
    args = argparse.Namespace(command=cmd, queue='queue', 
                              host='localhost', port=6547,
                              inputfile='/foo')
    if cmd == 'stage':
        # Have to mock the read_file call
        with patch('plurmy.orchestrate.read_file', lambda x: []) as rf:
            with patch(f'plurmy.orchestrate.{cmd}', create_autospec(lambda x, y, z:x)) as _func:
                orc.main(parser, args)
                _func.assert_called_once()
    else:
        # Have to mock the redis setup call
        def mock_redis_constructor(host=None, port=None, db=0):
            pass
    
        with patch('plurmy.orchestrate.StrictRedis', create_autospec(mock_redis_constructor)) as _conn:
            orc.setup_redis_connection('host', 12345)
            with patch(f'plurmy.orchestrate.{cmd}', create_autospec(lambda x, y, z:x)) as _func:
                orc.main(parser, args)
                _func.assert_called_once()


from setuptools import setup, find_packages

setup(
    name="plurmy",
    version="0.2.1",
    license='CC0-1.0',
    long_description="",
    packages=find_packages(),
    include_package_data=True,
    install_requires = [
     'fakeredis',
     'hiredis',
     'python>=3.7',
     'pytest',
     'pytest-cov',
     'redis',
     'redis-py' 
    ],
    entry_points = {
        'console_scripts': [
            'plurmy = plurmy.orchestrate:main',
        ]
    }
)



import subprocess
import datetime



def slurm_walltime_to_seconds(walltime: str) -> int:
    """
    Convert a slurm defined walltime in the form
    HH:MM:SS into a number of seconds.

    Parameters
    ----------
    walltime : str
            In the form HH:MM:SS

    Returns
    -------
    d : int
        The number of seconds the walltime represents

    Examples
    --------
    >>> walltime = '01:00:00'
    >>> sec = slurm_walltime_to_seconds(walltime)
    >>> sec
    3600
    """
    walltime = walltime.split(":")
    walltime = list(map(int, walltime))
    d = datetime.timedelta(hours=walltime[0], minutes=walltime[1], seconds=walltime[2])

    return d.seconds


class Slurm(object):
    """
    The Slurm class is the primary class that plurmy exposes. This class makes use of
    the __repr__ magic method to map class attributes to SBATCH directives. Therefore, the
    attribute names have a tight, implicit, mapping to the slurm directives.

    Attributes
    ----------
    command : str
              The command to be run, e.g., srun -n 10 myparallelscript.py

    job_name : str
               The name of the job to be run as viewed by commands like sacct and
               squeue

    time : str
           The job walltime in the form "DD:HH:MM:SS" or a truncated form, e.g., 
           "HH:MM:SS". 

    output : str
             The directory to output logs

    mem_per_cpu : int
                  The amount of memory, as an integer in MB to be allocated
                  per requested CPU

    nodes : str
            A comma separated list of node ids to be used by the job

    partition : str
                The partition name the job is submitted to

    ntasks : int
             The number of independent tasks (CPUs) to request
    """
    def __init__(
        self,
        command,
        job_name=None,
        time="01:00:00",
        output=None,
        mem_per_cpu=2048,
        nodes=None,
        partition=None,
        ntasks=1,
        account=None
    ):
        self.command = command
        self.job_name = job_name
        self.time = time
        self.output = output
        self.mem_per_cpu = mem_per_cpu
        self.nodes = nodes
        self.partition = partition
        self.ntasks = ntasks
        self.account = account

    @property
    def account(self):
        return getattr(self, "_account", None)

    @account.setter
    def account(self, account: str):
        self._account = account

    @property
    def job_name(self):
        return getattr(self, "_job_name", None)

    @job_name.setter
    def job_name(self, job_name: str):
        self._job_name = job_name

    @property
    def time(self):
        return getattr(self, "_time", None)

    @time.setter
    def time(self, time: str):
        self._time = time

    @property
    def output(self):
        return getattr(self, "_output", None)

    @output.setter
    def output(self, output: str):
        output = "{}{}".format(output, '.%A_%a.out')
        self._output = output

    @property
    def mem_per_cpu(self):
        return getattr(self, "_mem_per_cpu", 2048)

    @mem_per_cpu.setter
    def mem_per_cpu(self, mem: int):
        self._mem_per_cpu = mem

    @property
    def partition(self):
        return getattr(self, "_partition", None)

    @partition.setter
    def partition(self, partition: str):
        self._partition = partition

    @property
    def nodes(self):
        return getattr(self, "_nodes", None)

    @nodes.setter
    def nodes(self, nodes: str):
        self._nodes = nodes

    @property
    def command(self):
        return getattr(self, "_command")

    @command.setter
    def command(self, command: str):
        self._command = command

    @property
    def ntasks(self):
        return getattr(self, "_ntasks", 1)

    @ntasks.setter
    def ntasks(self, ntasks: int):
        self._ntasks = ntasks

    def _submit_one(self):
        proc = ["sbatch"]
        process = subprocess.Popen(proc, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        job_str = str(self)
        process.stdin.write(job_str.encode())
        out, err = process.communicate()
        if err:
            return False
        return job_str

    def submit(self, array=None, chunksize=1000, exclude=None):
        """Submits the slurm job.

        Parameters
        ----------
        array : str
                The Slurm formatted specification that describes array attributes.

        Returns
        -------
        job_str : str
            The string representation of the sbatch file submitted to slurm.


        Examples
        --------
        >>> slurm_job = Slurm('./foo.py')
        >>> slurm_job.submit()
        >>> slurm_job.submit("1-6")
        >>> slurm_job.submit("1-3, 8-9")
        """

        if array == None:
            return self._submit_one()
        else:
            job_strs = []
            arrays = []
            # Parse out the array max concurrent job string
            if "%" in array:
                array, step = array.split("%")
            else:
                step = None
            # Parse if different array job ids have been manually specified
            splits = array.split(",")
            for extent in splits:
                # Map the strings to ints for math
                try:
                    start, stop = list(map(int, extent.split("-")))
                except:
                    return self._submit_one()
                # Case where the total number of jobs is > the chunk size
                if stop - start + 1 > chunksize:
                    quot, rem = divmod(stop - start + 1, chunksize)
                    arrays = [f"1-{chunksize}"]
                    if step:
                        arrays[0] += f"%{step}"
                    arrays = arrays * quot
                    if rem:
                        remainder = f"1-{rem}"
                        if step:
                            remainder += f"%{step}"
                        arrays.append(remainder)
                # Total number of jobs is < the chunk size
                else:
                    inbounds = f"{start}-{stop}"
                    if step:
                        inbounds += f"%{step}"
                    arrays.append(inbounds)

            if exclude is not None:
                proc = ["sbatch", "--exclude={}".format(exclude)]
            else:
                proc = ["sbatch"]

            for array in arrays:
                proc = ["sbatch"]
                proc.extend(("--array", array))

                process = subprocess.Popen(
                    proc, stdin=subprocess.PIPE, stdout=subprocess.PIPE
                )
                job_str = str(self)
                job_strs.append(job_str)
                process.stdin.write(job_str.encode())
                out, err = process.communicate()
                if err:
                    return False
        return job_strs

    def __repr__(self):
        sbatch = "#SBATCH --{}{}"
        cmd = ["#!/bin/bash -l"]
        for k, v in vars(self).items():
            if k.startswith("__") or k == "_command":
                continue
            else:
                # Dict items are prefixed with '_' -- get rid of it
                k = k[1:] if k.startswith("_") else k
                if v is not None:
                    # Convert python style separators (_) to slurm separators.
                    k = k.replace("_", "-")
                    # Allow flag specification using empty string.  Necessary
                    # to support flags like '--test'
                    #  that have no arguments.
                    v = "{sep}{val}".format(sep="=" if str(v) else "", val=v)
                    cmd.append(sbatch.format(k, v))

        cmd.append(self.command)
        return "\n".join(str(s) for s in cmd)
